import luigi
import importlib
import os
import sys
from pathlib import Path
import pandas as pd
from .config_utilities import unfreeze
from .errors import OutputError

class Distillery(luigi.Task):
    """ Task that encapsulates analysis tasks and makes them executable
    inside the Datenraffinerie
    """
    name = luigi.Parameter(significant=True)
    python_module = luigi.Parameter(significant=True)
    daq = luigi.Parameter(significant=True)
    output_dir = luigi.Parameter(significant=True)
    parameters = luigi.DictParameter(significant=True)
    root_config_path = luigi.Parameter(significant=True)
    analysis_module_path = luigi.OptionalParameter(significant=True,
                                                   default=None)
    network_config = luigi.DictParameter(significant=False)
    loop = luigi.BoolParameter(significant=False)
    event_mode = luigi.BoolParameter(significant=False)

    def requires(self):
        from .valve_yard import ValveYard
        """ Determin which analysis needs to be run to produce
        the data for the analysis
        :returns: The acquisition procedure needed to produce the data
        """
        return ValveYard(self.root_config_path, self.daq,
                         os.path.dirname(self.output_dir),
                         self.analysis_module_path, self.network_config,
                         self.loop)

    def output(self):
        """ Define the files that are produced by the analysis
        :returns: dictionary with the keys 'summary', 'calibration'
            and 'plots', where 'summary's value is a string for a
            relative path to the summary, same as calibration is a
            relative path to the calibration yaml file
            and the value associated to 'plots' is a list of relative
            paths. All paths should be strings.
        """
        analysis = self.import_analysis(self.analysis_module_path,
                                        self.python_module)
        analysis_parameters = unfreeze(self.parameters)
        analysis = analysis(analysis_parameters)
        output = {}
        for key, paths in analysis.output().items():
            if key == 'plots':
                try:
                    if len(paths) == 0:
                        continue
                    output[key] = [luigi.LocalTarget(
                        (Path(self.output_dir) / path).resolve())
                        for path in paths]
                except TypeError as err:
                    raise OutputError('The plots output must be a list ' +
                                      'of paths. Use an empty list if ' +
                                      'no plots are generated') from err
            else:
                if paths is None:
                    continue
                output[key] = (Path(self.output_dir) / paths).resolve()
        return output

    def run(self):
        """ perform the analysis using the imported distillery
        :returns: TODO

        """
        # import the class definition
        analysis = self.import_analysis(self.analysis_module_path,
                                            self.python_module)
        # instantiate an analysis object from the imported analysis class
        analysis = analysis(unfreeze(self.parameters))

        # open and read the data
        if not self.event_mode:
            data = pd.read_hdf(self.input().path)
        else:
            data = pd.HDFStore(self.input().path, mode='r')
        analysis.run(data, self.output_dir)

    @staticmethod
    def import_analysis(distillery_path: str, name: str):
        """ Import the distillery for the analysis.

        :distillery_path: The path in which to find the distilleries
            module
        :name: The name of the distillery to load
        :returns: the distillery loaded into the local namespace
        """
        if distillery_path is not None:
            pathstr = str(Path(distillery_path).resolve())
            pythonpath_entry = os.path.split(pathstr)[0]
            module_name = os.path.split(pathstr)[1]
            sys.path.append(pythonpath_entry)
            i = importlib.import_module(module_name)
            distillery = getattr(i, name)
        else:
            import datenraffinerie_distilleries as distilleries
            distillery = getattr(distilleries, name)
        return distillery



"""
The ValveYard is a class in the datenraffinerie that is responsible
for parsing the user configuration and requiring the right scan or
analysis task and passing it all necessary values from their proper
functioning

Author: Alexander Becker (a.becker@cern.ch)
Date: 2021-12-16
"""
import os
from pathlib import Path
import luigi
import zmq
import yaml
from .scan import DataField, Fracker
from .distillery import Distillery
from . import config_utilities as cfu
from . import control_adapter as ctrl


class ValveYard(luigi.WrapperTask):
    root_config_file = luigi.Parameter(significant=True)
    procedure_label = luigi.Parameter(significant=True)
    output_dir = luigi.Parameter(significant=True)
    analysis_module_path = luigi.Parameter(significant=True)
    network_config = luigi.DictParameter(significant=True)
    loop = luigi.BoolParameter(significant=True)

    def output(self):
        return self.input()

    def requires(self):
        """ A wrapper that parses the configuration and starts the procedure
        with the corresponding procedure label

        :raises: ConfigFormatError if either the type of the configuration entry
            does not match the allowed ones or if the YAML is malformed
        :raises: DAQConfigError if the DAQ configuration is malformed
        """
        data_dir = Path(self.output_dir)
        if not data_dir.exists():
            os.makedirs(data_dir)
        procedures, workflows = cfu.parse_config_file(self.root_config_file)
        procedure_names = list(map(lambda p: p['name'], procedures))
        if self.procedure_label in procedure_names:
            procedure_index = procedure_names.index(self.procedure_label)
        else:
            raise ctrl.DAQConfigError(f"No '{self.procedure_label}' found in"
                                      " the config files")
        output_dir = data_dir / self.procedure_label
        if not output_dir.exists():
            os.makedirs(output_dir)
        procedure = procedures[procedure_index]
        if procedure['type'] == 'analysis':
            return Distillery(name=self.procedure_label,
                              python_module=procedure['python_module'],
                              daq=procedure['daq'],
                              output_dir=str(output_dir.resolve()),
                              parameters=procedure['parameters'],
                              root_config_path=str(
                                  Path(self.root_config_file).resolve()),
                              analysis_module_path=self.analysis_module_path,
                              network_config=self.network_config,
                              loop=self.loop,
                              event_mode=procedure['event_mode'])
        if procedure['type'] == 'daq':
            # the default values for the DAQ system and the target need to
            # be loaded on to the backend
            context = zmq.Context()
            socket = context.socket(zmq.REQ)
            socket.connect(
                    f"tcp://{self.network_config['daq_coordinator']['hostname']}:"
                    f"{self.network_config['daq_coordinator']['port']}")
            complete_default_config = {
                    'daq': procedure['daq_system_default_config'],
                    'target': procedure['target_power_on_default_config']}
            socket.send_string('load defaults;' +
                               yaml.safe_dump(complete_default_config))
            resp = socket.recv()
            if resp != b'defaults loaded':
                raise ctrl.DAQConfigError('Default config could not be loaded into the backend')
            if len(procedure['parameters']) == 1 and self.loop:
                return Fracker(identifier=0,
                               label=self.procedure_label,
                               output_dir=str(output_dir.resolve()),
                               output_format='hdf5',
                               scan_parameters=procedure['parameters'],
                               target_config=procedure['target_init_config'],
                               target_default_config=procedure['target_power_on_default_config'],
                               daq_system_config=procedure['daq_system_config'],
                               root_config_path=str(
                                   Path(self.root_config_file).resolve()),
                               calibration=procedure['calibration'],
                               analysis_module_path=self.analysis_module_path,
                               network_config=self.network_config,
                               loop=self.loop,
                               raw=procedure['raw'],
                               data_columns=procedure['data_columns']) # indicate if to produce event by event data data
            return DataField(identifier=0,
                             label=self.procedure_label,
                             output_dir=str(output_dir.resolve()),
                             output_format='hdf5',
                             scan_parameters=procedure['parameters'],
                             target_config=procedure['target_init_config'],
                             target_default_config=procedure['target_power_on_default_config'],
                             daq_system_config=procedure['daq_system_config'],
                             root_config_path=str(
                                 Path(self.root_config_file).resolve()),
                             calibration=procedure['calibration'],
                             analysis_module_path=self.analysis_module_path,
                             network_config=self.network_config,
                             loop=self.loop,
                             raw=procedure['raw'],
                             data_columns=procedure['data_columns']) # indicate if to produce event by event data data
        raise cfu.ConfigFormatError("The type of an entry must be either "
                                    "'daq' or 'analysis'")



import zmq
import yaml
import logging
from pathlib import Path
from . import control_adapter as ctrl

def coordinate_daq_access(network_config: dict):
    """ a function run in a separate process that coordinates the access
    to the daq-system and target system so that the access is serialized
    and the daq interface is stabilized towards the datenraffinerie

    :network_config: the hostnames and ports of the target, daq-client and daq-server
    :returns: Nothing

    """
    logging.basicConfig(filename='daq_coordinator.log',
                        encoding='utf-8',
                        level=logging.DEBUG)
    logger = logging.getLogger('daq_coordinator')
    logger.debug('creating zmq context and socket')
    context = zmq.Context()
    socket = context.socket(zmq.REP)
    socket_address = f"tcp://{network_config['daq_coordinator']['hostname']}:" +\
                     f"{network_config['daq_coordinator']['port']}"
    socket.bind(socket_address)
    logger.debug('bound to: %s' % socket_address)
    target = ctrl.TargetAdapter(network_config['target']['hostname'],
                                network_config['target']['port'])
    daq_system = ctrl.DAQSystem(network_config['server']['hostname'],
                                network_config['server']['port'],
                                network_config['client']['hostname'],
                                network_config['client']['port'])
    logger.debug('created the target and daq systems')
    while True:
        logger.debug('waiting for message')
        message = socket.recv()
        message_delimiter = message.find(b';')
        command = message[:message_delimiter]
        if command == b'measure':
            logger.debug('received message "measure"')
            if not target.has_default() and not daq_system.has_default():
                socket.send_string('error: no defaults loaded')
                continue
            config = yaml.safe_load(message[message_delimiter+1:].decode())
            logger.debug('Received Configuration:\n' + yaml.dump(config))
            daq_config = config['daq']
            target_config = config['target']
            target.configure(target_config, overlays_default=True)
            daq_system.configure(daq_config)
            measurement_data_path = Path('./measurement_data.raw')
            daq_system.take_data(measurement_data_path)
            with open(measurement_data_path, 'rb') as data_file:
                socket.send(data_file.read())
        elif command == b'load defaults':
            logger.debug('received message: %s' % command.decode())
            logger.debug('received default config:\n%s' %
                         message[message_delimiter+1:].decode())
            default_config = yaml.safe_load(
                    message[message_delimiter+1:].decode())
            target_default_config = default_config['target']
            daq_default_config = default_config['daq']
            target.load_default_config(target_default_config)
            daq_system.load_default_config(daq_default_config)
            socket.send_string('defaults loaded')
        else:
            logger.warning('invalid command message received: %s' %
                           command.decode())
            socket.send_string('invalid command')



import sys
import multiprocessing
import click
import luigi
import yaml
from .valve_yard import ValveYard
from .config_utilities import ConfigFormatError
from .control_adapter import DAQConfigError
from .daq_coordination import coordinate_daq_access


@click.command()
@click.argument('netcfg', type=click.File('r'),
                metavar='[network configuration file]')
@click.argument('config', type=click.Path(exists=True),
                metavar='[main configuration file]')
@click.argument('procedure', type=str,
                metavar='[Procedure to be run by the datenraffinerie]')
@click.argument('output', type=click.Path(),
                metavar='[data output directory]')
@click.option('-a', '--analysis_path', 'analysis_path', type=click.Path(exists=True),
              help='specify the path to the python module containing the analyses')
@click.option('-w', '--workers', 'workers', type=int, default=1,
              help='specify the amount of parallel tasks to be executed')
@click.option('-l', '--loop', 'loop', is_flag=True, default=False,
              help='tell the scan to run the measurements in a loop'
                   + ' instead of creating them in separate tasks')
def cli(netcfg, config, procedure, workers, output, analysis_path, loop):
    """ The command line interface to the datenraffinerie intended to
    be one of the primary interfaces for the users.
    """
    try:
        netcfg = yaml.safe_load(netcfg.read())
    except yaml.YAMLError as err:
        print('Error reading in the network config:\n'
              + str(err) + '\nexiting ..')
        sys.exit(1)
    daq_coordination_process = multiprocessing.Process(target=coordinate_daq_access, args=(netcfg, ))
    daq_coordination_process.start()
    try:
        run_result = luigi.build([ValveYard(
            click.format_filename(config),
            procedure,
            output,
            analysis_path,
            netcfg,
            loop='true' if loop else 'false')],
            local_scheduler=True,
            workers=workers,
        )
        print(run_result)
    except ConfigFormatError as err:
        print(err.message)
    except DAQConfigError as err:
        print("The Configuration of one of the executed"
              " DAQ procedures is malformed: ")
        print(err.message)
    except Exception as err:
        print("An error occured that was not properly caught")
        print(err)
    finally:
        daq_coordination_process.kill()
        daq_coordination_process.join()
        sys.exit(1)



"""
Module containing the classes that together constitute a measurement
and encapsulate the different steps needed to take a measurement using the
hexaboard
"""
from pathlib import Path
from functools import reduce
import os
import operator
import pandas as pd
import luigi
from luigi.parameter import ParameterVisibility
import subprocess
import yaml
import zmq
from uproot.exceptions import KeyInFileError
from . import config_utilities as cfu
from . import analysis_utilities as anu


class Calibration(luigi.Task):
    """
    fetches the Calibration for a daq procedure this is normally
    done by the
    """
    root_config_path = luigi.Parameter()
    calibration = luigi.Parameter()
    output_dir = luigi.Parameter()
    analysis_module_path = luigi.Parameter()
    loop = luigi.BoolParameter()

    def requires(self):
        from .valve_yard import ValveYard
        # if a calibration is needed then the delegate finding
        # the calibration and adding the subsequent tasks to the
        # to the ValveYard
        if self.calibration is not None:
            return ValveYard(self.root_config_path,
                             self.calibration,
                             os.path.dirname(str(Path(self.output_dir).resolve())),
                             str(Path(self.analysis_module_path).resolve()),
                             self.loop)

    def output(self):
        local_calib_path = Path(self.output_dir) / 'calibration.yaml'
        return luigi.LocalTarget(local_calib_path)

    def run(self):
        # figure out if there is a calibration that we need and if so create a
        # local copy so that we don't end up calling the valve yard multiple times
        if self.calibration is not None:
            with self.input()['calibration'].open('r') as calibration_file:
                with self.output().open('w') as local_calib_copy:
                    local_calib_copy.write(calibration_file.read())
        else:
            with self.output().open('w') as local_calib_copy:
                local_calib_copy.write('')


class DrillingRig(luigi.Task):
    """
    Task that unpacks the raw data into the desired data format
    also merges the yaml chip configuration with the reformatted
    data.
    """
    # configuration and connection to the target
    # (aka hexaboard/SingleROC tester)
    target_config = luigi.DictParameter(significant=False)
    target_default_config = luigi.DictParameter(significant=False)

    # configuration of the (daq) system
    daq_system_config = luigi.DictParameter(significant=False)

    # Directory that the data should be stored in
    output_dir = luigi.Parameter(significant=True)
    output_format = luigi.Parameter(significant=False)
    label = luigi.Parameter(significant=True)
    identifier = luigi.IntParameter(significant=True)

    # the path to the root config file so that the Configuration
    # task can call the valveyard if a calibration is required
    root_config_path = luigi.Parameter(True)
    # calibration if one is required
    calibration = luigi.OptionalParameter(significant=False)
    analysis_module_path = luigi.OptionalParameter(significant=False)
    network_config = luigi.DictParameter(significant=True)
    loop = luigi.BoolParameter(significant=False)
    raw = luigi.BoolParameter(significant=True)
    data_columns = luigi.ListParameter(significant=False)

    def requires(self):
        return Calibration(self.root_config_path,
                           self.calibration,
                           self.output_dir,
                           self.analysis_module_path,
                           self.loop)

    def output(self):
        """
        define the file that is to be produced by the unpacking step
        the identifier is used to make the file unique from the other
        unpacking steps
        """
        formatted_data_path = Path(self.output_dir) / \
            f'{self.label}_{self.identifier}.{self.output_format}'
        return luigi.LocalTarget(formatted_data_path.resolve())

    def run(self):
        # load the configurations
        target_config = cfu.unfreeze(self.target_config)
        daq_system_config = cfu.unfreeze(self.daq_system_config)
        power_on_default = cfu.unfreeze(self.target_default_config)

        # load the calibration
        if self.calibration is not None:
            with self.input().open('r') as calibration_file:
                calibration = yaml.safe_load(
                    calibration_file.read())
            # calculate the configuration to send to the backend
            target_config = cfu.update_dict(target_config,
                                            calibration)

        target_config = cfu.diff_dict(power_on_default,
                                      target_config)
        full_target_config = cfu.update_dict(power_on_default,
                                             target_config)
        complete_config = {'daq': daq_system_config,
                           'target': full_target_config}

        # create the config on disk
        output_config = os.path.splitext(self.output().path)[0] + '.yaml'
        config_string = yaml.safe_dump(complete_config)
        with open(output_config, 'w') as run_config:
            run_config.write(config_string)

        # send config to the backend and wait for the response
        context = zmq.Context()
        socket = context.socket(zmq.REQ)
        socket.connect(
                f"tcp://{self.network_config['daq_coordinator']['hostname']}:" +
                f"{self.network_config['daq_coordinator']['port']}")
        socket.send_string('measure;' + config_string)
        data = socket.recv()
        socket.close()
        context.term()
        raw_data_file_path = os.path.splitext(self.output().path)[0] + '.raw'

        # save the data in a file so that the unpacker can work with it
        with open(raw_data_file_path, 'wb') as raw_data_file:
            raw_data_file.write(data)

        data_file_path = os.path.splitext(self.output().path)[0] + '.root'

        result = anu.unpack_raw_data_into_root(raw_data_file_path,
                                  data_file_path,
                                  raw_data=self.raw)
        if result != 0:
            os.remove(raw_data_file_path)
            os.remove(data_file_path)
            raise ValueError(f"The unpacker failed for {raw_data_file_path}")

        # load the data from the unpacked root file and merge in the
        # data from the configuration for that run with the data
        output_path = Path(self.output().path)
        anu.reformat_data(data_file_path, output_path, complete_config, self.raw,
                          columns=self.data_columns)
        os.remove(data_file_path)


class DataField(luigi.Task):
    """
    A Scan over one parameter or over other scans

    The scan uses the base configuration as the state of the system
    and then modifies it by applying patches constructed from
    parameter/value pairs passed to the scan and then calling either
    the measurement task or a sub-scan with the patched configurations
    as their respective base configurations
    """
    # parameters describing the position of the parameters in the task
    # tree
    identifier = luigi.IntParameter(significant=True)

    # parameters describing to the type of measurement being taken
    # and the relevant information for the measurement/scan
    label = luigi.Parameter(significant=True)
    output_dir = luigi.Parameter(significant=True)
    output_format = luigi.Parameter(significant=False)
    scan_parameters = luigi.ListParameter(significant=False)

    # configuration of the target and daq system that is used to
    # perform the scan (This may be extended with an 'environment')
    target_config = luigi.DictParameter(significant=False)
    target_default_config = luigi.DictParameter(significant=False)
    daq_system_config = luigi.DictParameter(significant=False)

    root_config_path = luigi.Parameter(significant=True)
    # calibration if one is required
    calibration = luigi.OptionalParameter(significant=False,
                                          default=None)
    analysis_module_path = luigi.OptionalParameter(significant=False,
                                                   default=None)
    network_config = luigi.DictParameter(significant=False)
    loop = luigi.BoolParameter(significant=False)
    raw = luigi.BoolParameter(significant=False)
    data_columns = luigi.ListParameter(significant=False)

    supported_formats = ['hdf5']

    @property
    def priority(self):
        if self.loop and len(self.scan_parameters) == 1:
            return 10
        return 0

    def requires(self):
        """
        Determine the measurements that are required for this scan to proceed.

        The Scan class is a recursive task. For every parameter(dimension) that
        is specified by the parameters argument, the scan task requires a
        set of further scans, one per value of the values entry associated with
        the parameter that the current scan is to scan over, essentially
        creating the Cartesian product of all parameters specified.
        """
        required_tasks = []
        values = self.scan_parameters[0][1]
        parameter = self.scan_parameters[0][0]
        target_config = cfu.unfreeze(self.target_config)
        daq_system_config = cfu.unfreeze(self.daq_system_config)
        complete_config = {'target': target_config,
                           'daq': daq_system_config}

        # if there are more than one entry in the parameter list the scan still
        # has more than one dimension. So spawn more scan tasks for the lower
        # dimension
        if len(self.scan_parameters) > 1:
            # calculate the id of the task by multiplication of the length of
            # the dimensions still in the list
            task_id_offset = reduce(operator.mul,
                                    [len(param[1]) for param in
                                     self.scan_parameters[1:]])
            subscan_target_config = target_config
            subscan_daq_config = daq_system_config
            for i, value in enumerate(values):

                if type(value)==tuple:
                    value=list(value)

                patch = cfu.generate_patch(
                            parameter, value)
                complete_subscan_config = cfu.update_dict(complete_config, patch)
                subscan_daq_config = complete_subscan_config['daq']
                subscan_target_config = complete_subscan_config['target']
                if len(self.scan_parameters[1:]) == 1 and self.loop:
                    required_tasks.append(Fracker(self.identifier + 1 + task_id_offset * i,
                                                  self.label,
                                                  self.output_dir,
                                                  self.output_format,
                                                  self.scan_parameters[1:],
                                                  subscan_target_config,
                                                  self.target_default_config,
                                                  subscan_daq_config,
                                                  self.root_config_path,
                                                  self.calibration,
                                                  self.analysis_module_path,
                                                  self.network_config,
                                                  self.loop,
                                                  self.raw,
                                                  self.data_columns))
                else:
                    required_tasks.append(DataField(self.identifier + 1 + task_id_offset * i,
                                                    self.label,
                                                    self.output_dir,
                                                    self.output_format,
                                                    self.scan_parameters[1:],
                                                    subscan_target_config,
                                                    self.target_default_config,
                                                    subscan_daq_config,
                                                    self.root_config_path,
                                                    self.calibration,
                                                    self.analysis_module_path,
                                                    self.network_config,
                                                    self.loop,
                                                    self.raw,
                                                    self.data_columns))
        # The scan has reached the one dimensional case. Spawn a measurement
        # for every value that takes part in the scan
        else:
            if self.loop:
                return Calibration(self.root_config_path,
                                   self.calibration,
                                   self.output_dir,
                                   self.analysis_module_path,
                                   self.loop)

            for i, value in enumerate(values):
                patch = cfu.generate_patch(parameter, value)
                complete_config = cfu.update_dict(complete_config, patch)
                measurement_daq_config = complete_config['daq']
                measurement_target_config = complete_config['target']
                required_tasks.append(DrillingRig(measurement_target_config,
                                                  self.target_default_config,
                                                  measurement_daq_config,
                                                  self.output_dir,
                                                  self.output_format,
                                                  self.label,
                                                  self.identifier + i,
                                                  self.root_config_path,
                                                  self.calibration,
                                                  self.analysis_module_path,
                                                  self.network_config,
                                                  self.loop,
                                                  self.raw,
                                                  self.data_columns))
        return required_tasks

    def output(self):
        """
        generate the output file for the scan task

        If we are in the situation of being called by the fracker (first if condition)
        it is the job of the DataField to simply produce the raw files. It then also needs
        to figure out what files still need to be generated, as such it needs check what
        files have already been converted by the fracker. The fracker will fail and stall
        the rest of the luigi pipeline if it can't unpack the file. The user then needs to
        rerun the datenraffinerie
        """
        # we are being called by the fracker, so only produce the raw output files
        if len(self.scan_parameters) == 1 and self.loop:
            raw_files = []
            # pass the calibrated default config to the fracker
            raw_files.append(self.input())
            values = self.scan_parameters[0][1]
            for i, value in enumerate(values):
                base_file_name = f'{self.label}_{self.identifier + i}'
                raw_file_name = f'{base_file_name}.raw'
                fracked_file_name = f'{base_file_name}.hdf5'
                fracked_file_path = Path(self.output_dir) / fracked_file_name
                # if there already is a converted file we do not need to
                # acquire the data again
                if fracked_file_path.exists():
                    continue
                raw_file_path = Path(self.output_dir) / raw_file_name
                raw_files.append(luigi.LocalTarget(raw_file_path,
                                                   format=luigi.format.Nop))
            return raw_files

        # this task is not required by the fracker so we do the usual merge job
        if self.output_format in self.supported_formats:
            out_file = f'{self.label}_{self.identifier}_merged.{self.output_format}'
            raw_file_path = Path(self.output_dir) / out_file
            return luigi.LocalTarget(raw_file_path)
        raise KeyError("The output format for the scans needs to"
                       " one of the following:\n"
                       f"{self.supported_formats}")

    def run(self):
        """
        concatenate the files of a measurement together into a single file
        and write the merged data, or if the 'loop' parameter is set, it performs
        the measurements and lets the fracker handle the initial conversion into
        usable files if loop is set the fracker also does the merging at the
        end so in that case it is really 'just' there to acquire the data'.
        """

        # the fracker required us so we acquire the data and don't do any
        # further processing
        if self.loop and len(self.scan_parameters) == 1:
            # open the socket to the daq coordinator
            context = zmq.Context()
            socket = context.socket(zmq.REQ)
            socket.connect(
                f"tcp://{self.network_config['daq_coordinator']['hostname']}:"
                f"{self.network_config['daq_coordinator']['port']}")

            target_config = cfu.unfreeze(self.target_config)
            daq_system_config = cfu.unfreeze(self.daq_system_config)
            power_on_default = cfu.unfreeze(self.target_default_config)
            # load the calibration
            if self.calibration is not None:
                with self.input().open('r') as calibration_file:
                    calibration = yaml.safe_load(
                        calibration_file.read())
                # calculate the configuration to send to the backend
                target_config = cfu.update_dict(target_config, calibration)

            target_config = cfu.diff_dict(power_on_default,
                                          target_config)
            complete_config = {'daq': daq_system_config,
                               'target': target_config}

            # perform the scan
            values = self.scan_parameters[0][1]
            parameter = list(self.scan_parameters[0][0])
            output_files = self.output()[1:]
            output_configs = [os.path.splitext(of.path)[0] + '.yaml'
                              for of in output_files]
            for raw_file, output_config, value in\
                    zip(output_files, output_configs, values):
                if Path(raw_file.path).exists():
                    continue
                # patch the target config with the key for the current run
                # luigi might have converted an input list to a tuple
                if isinstance(value, tuple):
                    value = list(value)

                # generate the run configuration and write it to a file
                patch = cfu.generate_patch(parameter, value)
                complete_config = cfu.update_dict(complete_config, patch)
                with open(output_config, 'w') as run_config:
                    run_config.write(yaml.safe_dump(complete_config))

                # send the measurement command to the backend starting the
                # measurement
                socket.send_string('measure;'+yaml.safe_dump(complete_config))
                # wait for the data to return
                data = socket.recv()

                # save the data in a file so that the unpacker can work with it
                with raw_file.open('w') as raw_data_file:
                    raw_data_file.write(data)

            # close the connection to the daq coordinator
            # as the scan is now complete
            socket.close()
            context.term()

        # the measurements are being performed in the Measurement tasks
        # so the inputs are already unpacked hdf5 files and output is
        # the single merged file
        else:
            in_files = [data_file.path for data_file in self.input()]
            # merge the data together
            anu.merge_files(in_files, self.output().path, self.raw)


class Fracker(luigi.Task):
    """
    convert the format of the raw data into something that can be
    used by the distilleries
    """
    # parameters describing the position of the parameters in the task
    # tree
    identifier = luigi.IntParameter(significant=True)

    # parameters describing to the type of measurement being taken
    # and the relevant information for the measurement/scan
    label = luigi.Parameter(significant=True)
    output_dir = luigi.Parameter(significant=True)
    output_format = luigi.Parameter(significant=False)
    scan_parameters = luigi.ListParameter(significant=False)

    # configuration of the target and daq system that is used to
    # perform the scan (This may be extended with an 'environment')
    target_config = luigi.DictParameter(significant=False)
    target_default_config = luigi.DictParameter(significant=False)
    daq_system_config = luigi.DictParameter(significant=False)

    root_config_path = luigi.Parameter(significant=True)
    # calibration if one is required
    calibration = luigi.OptionalParameter(significant=False,
                                          default=None)
    analysis_module_path = luigi.OptionalParameter(significant=False,
                                                   default=None)
    network_config = luigi.DictParameter(significant=False)
    loop = luigi.BoolParameter(significant=False)
    raw = luigi.BoolParameter(significant=False)
    data_columns = luigi.ListParameter(significant=False)
    supported_formats = ['hdf5']

    def requires(self):
        return DataField(identifier=self.identifier,
                         label=self.label,
                         output_dir=self.output_dir,
                         output_format=self.output_format,
                         scan_parameters=self.scan_parameters,
                         target_config=self.target_config,
                         target_default_config=self.target_default_config,
                         daq_system_config=self.daq_system_config,
                         root_config_path=self.root_config_path,
                         calibration=self.calibration,
                         analysis_module_path=self.analysis_module_path,
                         network_config=self.network_config,
                         loop=self.loop,
                         raw=self.raw,
                         data_columns=self.data_columns)

    def output(self):
        """
        generate the output file for the scan task
        """
        if self.output_format in self.supported_formats:
            out_file = str(self.identifier) + '_merged.' + self.output_format
            output_path = Path(self.output_dir) / out_file
            return luigi.LocalTarget(output_path)
        raise KeyError("The output format for the scans needs to"
                       " one of the following:\n"
                       f"{self.supported_formats}")

    def run(self):
        # load the configurations
        target_config = cfu.unfreeze(self.target_config)
        daq_config = cfu.unfreeze(self.daq_system_config)
        power_on_default = cfu.unfreeze(self.target_default_config)
        # load the calibration
        if self.calibration is not None:
            with self.input()[0].open('r') as calibration_file:
                calibration = yaml.safe_load(
                    calibration_file.read())
            # calculate the configuration to send to the backend
            target_config = cfu.update_dict(target_config,
                                            calibration)

        target_config = cfu.update_dict(power_on_default,
                                        target_config)
        complete_config = {'daq': daq_config,
                           'target': target_config}

        for i, raw_file in enumerate(self.input()[1:]):
            data_file_base_name = os.path.splitext(raw_file.path)[0]
            unpacked_file_path = data_file_base_name + '.root'

            result = anu.unpack_raw_data_into_root(
                    raw_file.path,
                    unpacked_file_path,
                    raw_data=self.raw)
            if result != 0:
                os.remove(unpacked_file_path)

        # get the parameters to build the patch from
        values = self.scan_parameters[0][1]
        parameter = list(self.scan_parameters[0][0])
        expected_files = []
        for raw_file, value in zip(self.input()[1:], values):
            data_file_base_name = os.path.splitext(raw_file.path)[0]
            unpacked_file_path = Path(data_file_base_name + '.root')
            formatted_data_path = Path(data_file_base_name + '.hdf5')
            expected_files.append(formatted_data_path)
            # check that the unpacker was able to convert the data into root
            # format
            if not unpacked_file_path.exists():
                os.remove(raw_file.path)
                continue

            # load the data from the unpacked root file and merge in the
            # data from the configuration for that run with the data
            try:
                # calculate the patch that needs to be applied
                patch = cfu.generate_patch(parameter, value)
                complete_config = cfu.update_dict(complete_config, patch)
                anu.reformat_data(unpacked_file_path,
                                  formatted_data_path,
                                  complete_config,
                                  self.raw,
                                  columns=self.data_columns)
            except KeyInFileError:
                os.remove(unpacked_file_path.as_posix())
                continue
            except FileNotFoundError:
                continue
            os.remove(unpacked_file_path.as_posix())

        for formatted_file_path in expected_files:
            if not formatted_file_path.exists():
                raise ValueError('An unpacker failed, '
                                 'the datenraffinerie needs to be rerun')
        anu.merge_files(expected_files, self.output().path)



"""
Utilities for the use with the handling of the gathered data 
in the datenraffinerie.
"""
from . import config_utilities as cfu
from pathlib import Path
import os
import shutil
import pandas as pd
import numpy as np
import uproot
import tables
from numba import jit


class AnalysisError(Exception):
    def __init__(self, message):
        super().__init__(message)


def extract_data(rootfile: str, raw_data=False):
    """
    Extract the Data from the rootfile and put it
    into a Pandas dataframe
    """
    if raw_data is False:
        tree_name = 'runsummary/summary;1'
    else:
        tree_name = 'unpacker_data/hgcroc'

    with uproot.open(rootfile) as rfile:
        run_data = {}
        ttree = rfile[tree_name]
        for cname in ttree.keys():
            run_data[cname] = pd.Series(
                np.array(ttree[cname].array()),
                list(range(len(ttree[cname].array()))))
        return pd.DataFrame(run_data)


def unpack_raw_data_into_root(in_file_path, out_file_path, raw_data: bool=False):
    """
    'unpack' the data from the raw data gathered into a root file that
    can then be merged with the configuration into a large table
    """
    if raw_data:
        output_type = ' > /dev/null'
    else:
        output_type = ' -t unpacked'
    unpack_command = 'unpack'
    input_file = ' -i ' + str(in_file_path)
    output_command = ' -o ' + str(out_file_path)
    full_unpack_command = unpack_command + input_file + output_command\
        + output_type
    return os.system(full_unpack_command)


def create_empty_hdf_file(filename: str,
                          expectedrows: int,
                          compression: int = 1) -> tables.File:
    compression_filter = tables.Filters(complib='zlib', complevel=1)
    hdf_file = tables.open_file(filename, mode='w', filters=compression_filter)
    data = hdf_file.create_group('/', 'data')
    # create the attributes that are needed for pandas to
    # be able to read the hdf5 file as a dataframe
    data._v_attrs['axis0_variety'] = 'regular'
    data._v_attrs['axis1_variety'] = 'regular'
    data._v_attrs['encoding'] = 'UTF-8'
    data._v_attrs['errors'] = 'strict'
    data._v_attrs['ndim'] = 2
    data._v_attrs['nblocks'] = 0
    data._v_attrs['pandas_type'] = 'frame'
    data._v_attrs['pandas_version'] = '0.15.2'

    axis1 = hdf_file.create_earray(data, 'axis1', tables.IntAtom(),
                                   shape=(0,))
    axis1._v_attrs['kind'] = 'integer'
    axis1._v_attrs['name'] = 'rows'
    axis1._v_attrs['transposed'] = 1

    axis0 = hdf_file.create_earray(data, 'axis0', tables.StringAtom(50),
                                   shape=(0,))
    axis0._v_attrs['kind'] = 'string'
    axis0._v_attrs['name'] = 'columns'
    axis0._v_attrs['transposed'] = 1
    return hdf_file


def add_data_block(column_names: list, data_type, hdf_file: tables.File, group_name: str):
    hdf_group = hdf_file.get_node(f'/{group_name}')
    blockcount = hdf_group._v_attrs['nblocks']
    hdf_group._v_attrs[f'block{blockcount}_items_variety'] = 'regular'
    hdf_group._v_attrs['nblocks'] = blockcount + 1
    block_items = hdf_file.create_array(hdf_group,
                                        f'block{blockcount}_items',
                                        np.array(column_names))
    block_items._v_attrs['kind'] = 'string'
    block_items._v_attrs['name'] = 'N.'
    block_items._v_attrs['transposed'] = 1
    axis0 = hdf_file.get_node(f'/{group_name}/axis0')
    axis0.append(np.array(column_names))

    block_values = hdf_file.create_earray(hdf_group,
                                          f'block{blockcount}_values',
                                          data_type,
                                          shape=(0, len(column_names)))
    block_values._v_attrs['transposed'] = 1
    return block_values


def fill_block(file, group_name, block, data: np.ndarray):
    axis1 = file.get_node(f'/{group_name}/axis1')
    maxindex = len(axis1)
    block.append(np.array(data))
    if maxindex < len(block):
        axis1.append(np.arange(maxindex, len(block)))


def reformat_data(rootfile: str,
                  hdf_file: str,
                  complete_config: dict,
                  raw_data: bool,
                  chunk_length=50000,
                  columns: list = None):
    """ take in the unpacked root data and reformat into an hdf5 file.
    Also add in the configuration from the run.
    """
    hd_file = create_empty_hdf_file(hdf_file, 1000000)

    if raw_data:
        tree_name = 'unpacker_data/hgcroc'
        chan_id_branch_names = ['chip', 'channel', 'half']
    else:
        tree_name = 'runsummary/summary;1'
        chan_id_branch_names = ['chip', 'channel', 'channeltype']
    with uproot.open(rootfile) as rfile:
        # extract the data from the root file
        ttree = rfile[tree_name]
        if columns is not None:
            keys = list(filter(lambda x: x in columns, ttree.keys()))
        else:
            keys = list(ttree.keys())
        root_data = np.array([ttree[key].array() for key in keys]).transpose()
        blk_values = add_data_block(keys, tables.Int32Atom(), hd_file, 'data')
        fill_block(hd_file, 'data', blk_values, root_data)
        hd_file.flush()
        del root_data
        chip_chan_indices = np.array([ttree[key].array() for key in chan_id_branch_names]).transpose()

        # determin the size of the block that holds the configuration
        chan_config = roc_channel_to_dict(complete_config, 0, 0, 1)
        global_config = roc_channel_to_globals(complete_config, 0, 0, 1)
        l1a_config = l1a_generator_settings(complete_config)
        chan_config = cfu.update_dict(chan_config, global_config)
        chan_config = cfu.update_dict(chan_config, l1a_config)
        if columns is not None:
            config_keys = list(filter(lambda x: x in columns,
                                      chan_config.keys()))
        else:
            config_keys = list(chan_config.keys())
        config_blk = add_data_block(config_keys,
                                    tables.Int32Atom(),
                                    hd_file,
                                    'data')

        # merge in the configuration
        add_config_to_dataset(chip_chan_indices, hd_file, config_blk, complete_config,
                              chunk_length, raw_data, columns)
    hd_file.close()


def add_config_to_dataset(chip_chan_indices,
                          file,
                          dataset,
                          complete_config: dict,
                          chunklength: int,
                          raw_data: bool,
                          columns: list = None):
    """
    Add the configuration to the dataset

    extract the configuration for the chip/channel/type from the complete configuration
    for every row in the data and add it to the 'block1_values' EArray of the hdf5 file

    This function expects that the EArray has been created and all the neccesary metadata written
    it will simply add extract the data and append it to the EArray

    Parameters
    ----------
    chip_chan_indices : numpy.ndarray
        a 2D array where every row corresponds to [chip, channel, half_or_type]
    file: tables.File
        the file that the data should be written to
    dataset : tables.EArray
        the EArray that the data needs to be added to
    complete_config : dict
        the complete configuration of the chip at the time of the measurement
    chunklength : int
        the number of rows in a chunk of the hdf file this
    """
    chan_config = roc_channel_to_dict(complete_config, 0, 0, 1)
    global_config = roc_channel_to_globals(complete_config,
                                           0, 0, 1)
    l1a_config = l1a_generator_settings(complete_config)
    chan_config.update(global_config)
    chan_config.update(l1a_config)
    if columns is not None:
        config_keys = list(filter(lambda x: x in columns,
                                  chan_config.keys()))
    else:
        config_keys = list(chan_config.keys())
    chunks = int(len(chip_chan_indices) / chunklength)
    for i in range(chunks):
        chunk_array = np.zeros(shape=(chunklength, len(config_keys)))
        for j in range(chunklength):
            row = chip_chan_indices[i*chunklength+j]
            chip, chan, half_or_type = row[0], row[1], row[2]
            if raw_data:
                chip, chan, chan_type = \
                 compute_channel_type_from_event_data(chip, chan, half_or_type)
            else:
                chan_type = half_or_type
            chan_config.update(roc_channel_to_dict(complete_config,
                                                   chip, chan, chan_type))
            chan_config.update(roc_channel_to_globals(complete_config,
                                                      chip, chan, chan_type))
            if columns is not None:
                chunk_array[j] = np.array(list(map(lambda x: x[1],
                                                   filter(lambda x: x[0] in columns,
                                                          chan_config.items()))))
            else:
                chunk_array[j] = np.array(list(chan_config.values()))
        fill_block(file, 'data', dataset, chunk_array)
        file.flush()
    # add to the configuration to the last of the items that dont fully
    # fill up a chunk
    remaining_items = chip_chan_indices[chunks*chunklength:]
    chunk_array = np.zeros(shape=(len(remaining_items),
                                  len(config_keys)))
    for j, row in enumerate(remaining_items):
        chip, chan, half_or_type = row[0], row[1], row[2]
        if raw_data:
            chip, chan, chan_type = \
                compute_channel_type_from_event_data(chip, chan, half_or_type)
        else:
            chan_type = half_or_type
        chan_config.update(roc_channel_to_dict(complete_config,
                                               chip, chan, chan_type))
        chan_config.update(roc_channel_to_globals(complete_config,
                                                  chip, chan, chan_type))
        if columns is not None:
            chunk_array[j] = np.array(list(map(lambda x: x[1],
                                               filter(lambda x: x[0] in columns,
                                                      chan_config.items()))))
        else:
            chunk_array[j] = np.array(list(chan_config.values()))
    fill_block(file, 'data', dataset, chunk_array)


def merge_files(in_files: list, out_file: str, group_name: str='data'):
    """
    Merge the files of the different runs into a single file containing all
    the data
    """
    start_file = in_files[0]
    in_files = [tables.open_file(in_f, 'r') for in_f in in_files[1:]]
    # compression_filter = tables.Filters(complib='zlib', complevel=5)
    # copy the first of the input files to become the merged output
    shutil.copy(start_file, out_file)
    hd_file = tables.open_file(out_file, mode='r+')
    axis0 = hd_file.root.data.axis0
    nblocks = hd_file.root.data._v_attrs['nblocks']
    chunksize = 50000
    for in_f in in_files:
        in_ax0 = in_f.root.data.axis0
        inblocks = in_f.root.data._v_attrs['nblocks']
        if nblocks != inblocks:
            raise AnalysisError('the number of Blocks must be '
                                'Identical to merge the files')
        for elem, in_elem in zip(axis0, in_ax0):
            if elem != in_elem:
                raise AnalysisError('the columns of the files '
                                    'to be merged must match')

        for i in range(inblocks):
            out_block_items = getattr(hd_file.root.data, f'block{i}_items')
            out_block_values = getattr(hd_file.root.data, f'block{i}_values')
            in_block_items = getattr(in_f.root.data, f'block{i}_items')
            in_block_values = getattr(in_f.root.data, f'block{i}_values')
            for elem, in_elem in zip(out_block_items, in_block_items):
                if elem != in_elem:
                    raise AnalysisError('The items in the '
                                        f'block {i} don\'t match')
            # append the data from the input block to the output block
            chunks = len(in_block_values) // chunksize
            for chunk in range(chunks):
                start = chunk * chunksize
                stop = start + chunksize
                in_data = in_block_values.read(start, stop)
                fill_block(hd_file, 'data', out_block_values, in_data)
            # append the data that did not fill an entire chunk
            start = chunks * chunksize
            stop = len(in_block_values)
            in_data = in_block_values.read(start, stop)
            fill_block(hd_file, 'data', out_block_values, in_data)
        in_f.close()
        hd_file.flush()
    hd_file.close()


@jit(forceobj=True)
def l1a_generator_settings(complete_config: dict) -> dict:
    """add the config information of the chip half that corresponds to the particular
    channel

    :measurement_data: data measured from the rocs
    :complete_config: configuration of the rocs at the time of measurement.
        the half wise parameters of the rocs from this config will be added to every
        channel of the corresponding half in the `measurement_data`
    :returns: the dataframe measurement_data with added columns for the half wise
              parameters
    """
    l1a_config = {}
    l1a_generators = complete_config['daq']["server"]["l1a_generator_settings"]
    for l1a_generator in l1a_generators:
        name = l1a_generator["name"]
        for key,value in l1a_generator.items():
            if key=="name": continue
            if key=="flavor": continue
            if key=="followMode": continue
            output_key = name + "_" + str(key)
            l1a_config[output_key] = value
    return l1a_config


@jit(nopython=True)
def compute_channel_type_from_event_data(chip: int, channel: int, half: int):
    if channel <= 35:
        out_channel = channel * (half + 1)
        out_type = 0
    if channel == 36:
        out_channel = half
        out_type = 1
    if channel > 36:
        out_channel = channel - 37 + (half * 2)
        out_type = 100
    return chip, out_channel, out_type


@jit(nopython=False, forceobj=True)
def roc_channel_to_dict(complete_config: dict, chip_id: int, channel_id: int, channel_type: int) -> dict:
    """Map a channel identifier to the correct part of the config

    :chip_id: chip_id from the measurement in range 0,1,2 for LD hexaboard
    :channel_id: the channel number of the channel
    :channel_type: the channel type from the measurement
    :complete_config: the complete config of the chip
    :returns: TODO

    """
    target_subconfig = complete_config['target']
    id_map = {0: 'roc_s0', 1: 'roc_s1', 2: 'roc_s2'}
    channel_type_map = {0: 'ch', 1: 'calib', 100: 'cm'}
    return target_subconfig[id_map[int(chip_id)]]\
        [channel_type_map[int(channel_type)]][int(channel_id)]


def roc_channel_to_globals(complete_config: dict, chip_id: int, channel_id: int, channel_type: int) -> dict:
    """get the chip-half wise configuration from the 

    :chip_id: TODO
    :channel_id: TODO
    :channel_type: the channel type either a normal channel.
                   calibration channel or common mode channel
    :complete_config: The complete configuration of the chip at time of measurement
    :returns: a dictionary of all global setting for the given channel

    """
    target_subconfig = complete_config['target']
    id_map = {0: 'roc_s0', 1: 'roc_s1', 2: 'roc_s2'}
    channel_type_map = {0: 'ch', 1: 'calib', 100: 'cm'}
    half_wise_keys = ['DigitalHalf', 'GlobalAnalog', 'MasterTdc', 'ReferenceVoltage']
    global_keys = ['Top']
    channel_type = channel_type_map[channel_type]
    if channel_type == 'ch':
        chip_half = 0 if channel_id < 36 else 1
    if channel_type == 'cm':
        chip_half = 0 if channel_id < 2 else 1
    if channel_type == 'calib':
        chip_half = channel_id
    roc_config = target_subconfig[id_map[chip_id]]
    result = {}
    for hw_key in half_wise_keys:
        result.update(roc_config[hw_key][chip_half])
    for gl_key in global_keys:
        result.update(roc_config[gl_key][0])
    return result

# this stuff we define for the tests
data_columns = ['chip', 'channel', 'channeltype',
                    'adc_median', 'adc_iqr', 'tot_median',
                    'tot_iqr', 'toa_median', 'toa_iqr', 'adc_mean',
                    'adc_stdd', 'tot_mean',
                    'tot_stdd', 'toa_mean', 'toa_stdd', 'tot_efficiency',
                    'tot_efficiency_error', 'toa_efficiency',
                    'toa_efficiency_error']
expected_columns = [
    'Adc_pedestal', 'Channel_off', 'DAC_CAL_CTDC_TOA', 'DAC_CAL_CTDC_TOT',
    'DAC_CAL_FTDC_TOA', 'DAC_CAL_FTDC_TOT', 'DIS_TDC', 'ExtData',
    'HZ_inv', 'HZ_noinv', 'HighRange', 'IN_FTDC_ENCODER_TOA',
    'IN_FTDC_ENCODER_TOT', 'Inputdac', 'LowRange', 'mask_AlignBuffer',
    'mask_adc', 'mask_toa', 'mask_tot', 'probe_inv',
    'probe_noinv', 'probe_pa', 'probe_toa', 'probe_tot',
    'sel_trig_toa', 'sel_trig_tot', 'trim_inv', 'trim_toa',
    'trim_tot', 'Adc_TH', 'Bx_offset', 'CalibrationSC',
    'ClrAdcTot_trig', 'IdleFrame', 'L1Offset', 'MultFactor',
    'SC_testRAM', 'SelTC4', 'Tot_P0', 'Tot_P1',
    'Tot_P2', 'Tot ', 'Tot_P_Add',
    'Tot_TH0', 'Tot_TH1', 'Tot_TH2', 'Tot_TH3',
    'sc_testRAM', 'Cf', 'Cf_comp', 'Clr_ADC',
    'Clr_ShaperTail', 'Delay40', 'Delay65', 'Delay87',
    'Delay9', 'En_hyst_tot', 'Ibi_inv', 'Ibi_inv_buf',
    'Ibi_noinv', 'Ibi_noinv_buf', 'Ibi_sk', 'Ibo_inv',
    'Ibo_inv_buf', 'Ibo_no ', 'Ibo_noinv_buf',
    'Ibo_sk', 'ON ', 'ON_ref_adc',
    'ON_rtr', 'ON_toa', 'ON_tot', 'Rc',
    'Rf', 'S_inv', 'S_inv_buf',
    'S_noinv', 'S_noinv_buf', 'S_sk', 'SelExtADC',
    'SelRisingEdge', 'dac_pol', 'gain_tot', 'neg',
    'pol_trig_toa', 'range_indac', 'range_inv', 'range_tot',
    'ref_adc', 'trim_vbi_pa', 'trim_vbo_pa', 'BIAS_CAL_DAC_CTDC_P_D',
    'BIAS_CAL_DAC_CTDC_P_EN', 'BIAS_FOLLOWER_CAL_P_CTDC_EN',
    'BIAS_FOLLOWER_CAL_P_D', 'BIAS_FOLLOWER_CAL_P_FTDC_D',
    'BIAS_FOLLOWER_CAL_P_FTDC_EN', 'BIAS_I_CTDC_D',
    'BIAS_I_FTDC_D', 'CALIB_CHANNEL_DLL',
    'CTDC_CALIB_FREQUENCY', 'CTRL_IN_REF_CTDC_P_D',
    'CTRL_IN_REF_CTDC_P_EN', 'CTRL_IN_REF_FTDC_P_D',
    'CTRL_IN_REF_FTDC_P_EN', 'CTRL_IN_SIG_CTDC_P_D',
    'CTRL_IN_SIG_CTDC_P_EN', 'CTRL_IN_SIG_FTDC_P_D',
    'CTRL_IN_SIG_FTDC_P_EN', 'EN_MASTER_CTDC_DLL',
    'EN_MASTER_CTDC_VOUT_INIT', 'EN_MASTER_FTDC_DLL',
    'EN_MASTER_FTDC_VOUT_INIT', 'EN_REF_BG',
    'FOLLOWER_CTDC_EN', 'FOLLOWER_FTDC_EN',
    'FTDC_CALIB_FREQUENCY', 'GLOBAL_DISABLE_TOT_LIMIT',
    'GLOBAL_EN_BUFFER_CTDC', 'GLOBAL_EN_BUFFER_FTDC',
    'GLOBAL_EN_TOT_PRIORITY', 'GLOBAL_EN_TUNE_GAIN_DAC',
    'GLOBAL_FORCE_EN_CLK', 'GLOBAL_FORCE_EN_OUTPUT_DATA',
    'GLOBAL_FORCE_EN_TOT', 'GLOBAL_INIT_DAC_B_CTDC',
    'GLOBAL_LATENCY_TIME', 'GLOBAL_MODE_FTDC_TOA',
    'GLOBAL_MODE_NO_TOT_SUB', 'GLOBAL_MODE_TOA_DIRECT_OUTPUT',
    'GLOBAL_SEU_TIME_OUT', 'GLOBAL_TA_SELECT_GAIN_TOA',
    'GLOBAL_TA_SELECT_GAIN_TOT', 'INV_FRONT_40MHZ',
    'START_COUNTER', 'VD_CTDC_N_D',
    'VD_CTDC_N_DAC_EN', 'VD_CTDC_N_FORCE_MAX',
    'VD_CTDC_P_D', 'VD_CTDC_P_DAC_EN',
    'VD_FTDC_N_D', 'VD_FTDC_N_DAC_EN',
    'VD_FTDC_N_FORCE_MAX', 'VD_FTDC_P_D',
    'VD_FTDC_P_DAC_EN', 'sel_clk_rcg', 'Calib', 'ExtCtest',
    'IntCtest', 'Inv_vref', 'Noinv_vref', 'ON_dac',
    'Refi', 'Toa_vref', 'Tot_vref', 'Vbg_1v',
    'probe_dc', 'probe_dc1', 'probe_dc2', 'BIAS_I_PLL_D',
    'DIV_PLL', 'EN', 'EN_HIGH_CAPA', 'EN_LOCK_CONTROL',
    'EN_PLL', 'EN_PhaseShift', 'EN_RCG', 'EN_REF_BG',
    'EN_probe_pll', 'ENpE', 'ERROR_LIMIT_SC', 'EdgeSel_T1',
    'FOLLOWER_PLL_EN', 'INIT_D', 'INIT_DAC_EN', 'Pll_Locked_sc',
    'PreL1AOffset', 'RunL', 'RunR', 'S',
    'TestMode', 'VOUT_INIT_EN', 'VOUT_INIT_EXT_D', 'VOUT_INIT_EXT_EN',
    'b_in', 'b_out', 'err_countL', 'err_countR',
    'fc_error_count', 'in_inv_cmd_rx', 'lock_count', 'n_counter_rst',
    'phase_ck', 'phase_strobe', 'rcg_gain', 'sel_40M_ext',
    'sel_error', 'sel_lock', 'sel_strobe_ext', 'srout',
    'statusL', 'statusR', 'Tot_P3', 'Ibo_noinv', 'ON_pa'
    ]

daq_columns = ['Bx_trigger',
               'A_enable',
               'B_enable',
               'C_enable',
               'D_enable',
               'A_BX',
               'B_BX',
               'C_BX',
               'D_BX',
               'A_length',
               'B_length',
               'C_length',
               'D_length',
               'A_prescale',
               'B_prescale',
               'C_prescale',
               'D_prescale',
               ]

def test_extract_data():
    test_root_path = Path('../../tests/data/test_run_1.root')
    unpack_raw_data_into_root('../../tests/data/test_run_1.raw', test_root_path)
    frame = extract_data(test_root_path.resolve())
    for col in data_columns:
        _ = frame[col]


def test_reformat_data():
    """ test that all the channel wise parameters appear as a
    column in the dataframe
    """
    from . import config_utilities as cfu
    
    raw_data_path = Path('../../tests/data/test_run_1.raw')
    root_data_path = Path('./test_run_1.root')
    unpack_raw_data_into_root(raw_data_path, root_data_path, False)
    test_hdf_path = Path('./test_run_1.hdf5')

    # load the configuration of the target for the run
    configuration = cfu.load_configuration('../../tests/configuration/defaults/V3LDHexaboard-poweron-default.yaml')
    overlay = cfu.load_configuration('../../tests/data/test_run_1.yaml')

    # load the daq config and merge everything together
    daq_config = cfu.load_configuration('../../tests/configuration/defaults/daq-system-config.yaml')
    configuration = cfu.update_dict(configuration, overlay)
    configuration = {'target': configuration, 'daq': daq_config}

    # run the function under test
    reformat_data(root_data_path, test_hdf_path, configuration, False)

    # check that the columns from the config show up in the dataframe
    test_df = pd.read_hdf(test_hdf_path)
    for col in test_df.columns:
        assert col in expected_columns + data_columns + daq_columns
    os.remove(test_hdf_path)

    filtered_columns = ['toa_stdd', 'chip', 'channel', 'channeltype', 'toa_mean',
                        'A_BX', 'HighRange', 'LowRange']
    filter_columns_not_in_data = ['this', 'that']
    reformat_data(root_data_path, test_hdf_path, configuration, False,
                  1000, filtered_columns + filter_columns_not_in_data)
    test_df = pd.read_hdf(test_hdf_path)
    for col in filtered_columns:
        assert col in test_df.columns
    assert test_df.shape[1] == len(filtered_columns)
    os.remove(test_hdf_path)
    os.remove(root_data_path)


def test_merge_files():
    import glob
    from . import config_utilities as cfu
    default_target_config = cfu.load_configuration(
       '../../tests/configuration/defaults/V3LDHexaboard-poweron-default.yaml')
    daq_system_config = cfu.load_configuration(
       '../../tests/configuration/defaults/daq-system-config.yaml')
    default_config = {'target': default_target_config,
                      'daq': daq_system_config}
    raw_files = glob.glob('../../tests/data/*.raw')
    base_names = [os.path.splitext(raw_file)[0] for raw_file in raw_files]
    raw_files = [Path(raw_f) for raw_f in raw_files]
    config_names = [Path(bn + '.yaml') for bn in base_names]
    root_names = [Path(bn + '.root') for bn in base_names]
    dataframe_names = [Path(bn + '.hdf5') for bn in base_names]

    total_rows = 0
    columns = len(data_columns + expected_columns + daq_columns)
    for rawf, rootf, conff, dataff in \
            zip(raw_files, root_names, config_names, dataframe_names):
        unpack_raw_data_into_root(rawf, rootf, False)
        run_config = cfu.load_configuration(conff)
        run_config = cfu.update_dict(default_config, run_config)
        reformat_data(str(rootf.absolute()),
                      str(dataff.absolute()),
                      run_config, False, 1000)
        df_shape = pd.read_hdf(dataff).shape
        total_rows += df_shape[0]
        # assert df_shape[1] == columns

    out_file = 'merged.hdf5'
    merge_files(dataframe_names, out_file, 'data')
    final_df_shape = pd.read_hdf(out_file).shape
    assert final_df_shape[0] == total_rows
    # assert final_df_shape[1] == columns
    for rootf in root_names:
        if rootf.exists():
            os.remove(rootf)
    for df_path in dataframe_names:
        if df_path.exists():
            os.remove(df_path)
    os.remove(out_file)



"""
Module containing the utilities to parse and properly prepare the configuration
provided by the user

The functions here are testable by calling pytest config_utilities.py
The tests can also be used as a guide on how to use the functions and
what to expect of their behaviour
"""
from pathlib import Path
import os
from copy import deepcopy
from typing import Union
from luigi.freezing import FrozenOrderedDict
import yaml
import jinja2


class ConfigPatchError(Exception):
    def __init__(self, message):
        self.message = message


class ConfigFormatError(Exception):
    def __init__(self, message):
        self.message = message


def load_configuration(config_path):
    """
    load the configuration dictionary from a yaml file

    :raises: ConfigFormatError if the input cannot be parsed by the yaml
    parser
    """
    with open(config_path, 'r', encoding='utf-8') as config_file:
        try:
            return yaml.safe_load(config_file.read())
        except yaml.YAMLError as e:
            raise ConfigFormatError("unable to load yaml") from e


def unfreeze(config):
    """
    unfreeze the dictionary that is frozen to serialize it
    and send it from one luigi task to another
    """
    try:
        unfrozen_dict = {}
        for key in config.keys():
            unfrozen_dict[key] = unfreeze(config[key])
        return unfrozen_dict
    except AttributeError:
        try:
            unfrozen_list = []
            if not isinstance(config, str):
                for elem in config:
                    unfrozen_list.append(unfreeze(elem))
                return unfrozen_list
            return config
        except TypeError:
            return config


def update_dict(original: dict, update: dict, offset: bool = False,
                in_place: bool = False):
    """
    update one multi level dictionary with another. If a key does not
    exist in the original, it is created. the updated dictionary is returned

    Arguments:
    original: the unaltered dictionary

    update: the dictionary that will either overwrite or extend the original

    offset: If True, the values in the update dictionary will be interpreted as
            offsets to the values in the original dict. The resulting value wil
            be original + update (this will be string concatenation if both the
            original and update values are strings

    Returns:
    updated_dict: the extended/updated dict where the values that do not appear
                  in the update dict are from the original dict, the keys that
                  do appear in the update dict overwrite/extend the values in
                  the original
    Raises:
    TypeError: If offset is set to true and the types don't match the addition
               of the values will cause a TypeError to be raised

    ConfigPatchError: If lists are updated the length of the lists needs to match
               otherwise a error is raised
    """

    if in_place is True:
        result = original
    else:
        result = deepcopy(original)
    for update_key in update.keys():
        if update_key not in result.keys():
            if in_place is True:
                result[update_key] = update[update_key]
            else:
                result[update_key] = deepcopy(update[update_key])
            continue
        if not isinstance(result[update_key], type(update[update_key])):

            if in_place is True:
                result[update_key] = update[update_key]
            else:
                result[update_key] = deepcopy(update[update_key])
            continue
        if isinstance(result[update_key], dict):
            result[update_key] = update_dict(result[update_key],
                                             update[update_key],
                                             offset=offset,
                                             in_place=in_place)
        elif isinstance(result[update_key], list) or\
             isinstance(result[update_key], tuple):
            if len(result[update_key]) != len(update[update_key]):
                raise ConfigPatchError(
                    'If a list is a value of the dict the list'
                    ' lengths in the original and update need to'
                    ' match')
            patch = []
            for i, (orig_elem, update_elem) in enumerate(
                    zip(result[update_key], update[update_key])):
                if isinstance(orig_elem, dict):
                    patch.append(update_dict(result[update_key][i],
                                             update_elem,
                                             offset=offset,
                                             in_place=in_place))
                else:
                    if in_place is True:
                        patch.append(update_elem)
                    else:
                        patch.append(deepcopy(update_elem))
            if isinstance(result[update_key], tuple):
                patch = tuple(patch)
            result[update_key] = patch
        else:
            if offset is True:
                result[update_key] += update[update_key]
            else:
                if in_place is True:
                    result[update_key] = update[update_key]
                else:
                    result[update_key] = deepcopy(update[update_key])
    return result


def patch_configuration(config: dict, patch: dict):
    """
    currently only a wrapper function, if tests need to be implemented,
    then this is done here. Also the correct part of the cofiguration
    may be selected
    """
    return update_dict(config, patch)


def generate_patch(keys: Union[list, str], value):
    """
    the Scan task receives a list of keys and values and needs
    to generate a patch from one of the values of the list
    and all the keys. This function does that
    """
    # if the key is in fact a template for a yaml file
    if isinstance(keys, str):
        patch = yaml.safe_load(jinja2.Template(keys).render(value=value))
        if len(patch.keys()) != 1 and \
                (patch.keys()[0] != 'target' or patch.keys()[0] != 'daq'):
            patch = {'target': patch}
        return patch

    # this is needed to work with luigi as it turns stuffinto
    # tuples
    keys = list(keys)
    # here we need to insert 'target' if the key 'target' or 'daq' is not present
    # as the top level key to extend the scannable parameters to both the daq and target
    # configuration
    if keys[0] not in ['target', 'daq']:
        keys.insert(0, 'target')

    # now the rest of the generation can run as usual
    if len(keys) == 0:
        return {}
    keys.reverse()
    current_root = value
    for key in keys:
        level = {}
        if isinstance(key, list) or isinstance(key, tuple):
            for subkey in key:
                level[subkey] = deepcopy(current_root)
        else:
            level[key] = current_root
        current_root = level
    return current_root


def diff_dict(d1: dict, d2: dict) -> dict:
    """
    create the diff of two dictionaries.

    Create a dictionary containing all the keys that differ between
    d1 and d2 or that exist in d2 but not in d1. The value of the
    differing key will be the value found in d2.

    Example:
        With the inputs being
        d1 = {'a': 1, 'b': {'c': 2, 'f': 4}, 'e': 3}
        d2 = {'a': 2, 'b': {'c': 3, 'f': 4}, 'e': 3, 'g': 4}

        the resulting diff will be
        diff = {'a': 2, 'b': {'c':3}, 'g': 4}

        showing that the keys 'a' and 'b:c' are different and that the
        value in the differing d2 is 2 for a and 3 for 'b:c' respectively.
        As there is an entry in d2 that is not in d1 it will appear in the
        diff, however if there is an entry in d1 that is not in d2 it will
        not appear.

    This function is intended to find the patch that where applied on d1 to
    create d2. This may provide some motivation for the way that the function
    handles the different dictionaries
    """
    diff = {}
    for key2, value2 in d2.items():
        if key2 not in d1.keys():
            diff[key2] = value2
        elif isinstance(d1[key2], dict) and isinstance(value2, dict):
            potential_diff = diff_dict(d1[key2], value2)
            if potential_diff is not None:
                diff[key2] = potential_diff
        elif value2 != d1[key2]:
            diff[key2] = value2
    if diff == {}:
        return None
    return diff


def parse_scan_config(scan_config: dict, path: str) -> tuple:
    """
    parse the different entries of the dictionary to create
    the configuration that can be passed to the higher_order_scan function

    The function returns a tuple that contains some of the information that
    needs to be passed to the scan task

    The scan configuration allows it to override the parameters of a default
    daq-system configuration with ones specified in the sever_override and
    client_override section

    Arguments:
        scan_config: Dictionary that is the configuration for a single
                     scan/measurement
        path: Path to the file that contained the scan_config dict in yaml
              format this is needed to be able to find the daq-system default
              configuration
    Returns:
        the tuple of scan parameters, scan_requirement, scan_name and
        daq_system configuration.
    """
    daq_label = scan_config['name']
    scan_config_path = Path(path)
    scan_file_dir = Path(os.path.dirname(scan_config_path))

    # check if any calibrations need to be performed before taking data
    try:
        calibration = scan_config['calibration']
    except KeyError:
        calibration = None

    try:
        raw_data = scan_config['event_mode']
        if not isinstance(raw_data, bool):
            raise ConfigFormatError("The event_data field needs to contain a"
                                    " boolean")
    except KeyError:
        raw_data = False
    # parse the daq-system configuration
    # build the path for the default config
    try:
        daq_settings = scan_config['daq_settings']
    except KeyError as err:
        raise ConfigFormatError("A daq Task must specify a 'daq_settings'"
                                " section that at least contains the "
                                "'default' key pointing"
                                " to the default system config used") from err
    try:
        default_daq_settings_path = scan_file_dir / \
            Path(daq_settings['default'])
    except KeyError as err:
        raise ConfigFormatError("A daq Task must specify a 'daq_settings'"
                                " section that at least contains the "
                                "'default' key pointing"
                                " to the default system config used") from err
    # load the default config
    try:
        daq_system_config = load_configuration(
                default_daq_settings_path.resolve())
        daq_system_default_config = daq_system_config
    except FileNotFoundError as err:
        raise ConfigFormatError(f"The path specified for the default config"
                                f"of the daq_settings in {daq_label} does"
                                " not exist") from err
    # apply the override parameters for the server and the client
    for override_prefix in ['server', 'client']:
        override_key = override_prefix+'_override'
        try:
            override_config = daq_settings[override_key]
        except KeyError:
            override_config = None
        if isinstance(override_config, dict):
            override_config = {override_prefix: override_config}
            daq_system_config = update_dict(daq_system_config,
                                            override_config)
        # if the override entry is empty then don't do anything
        elif override_config is not None:
            raise ConfigFormatError(f"The {override_key} section in the daq"
                                    f"task {daq_label} must be a yaml dict")

    # load the power on and initial config for the target
    try:
        target_config = scan_config['target_settings']
    except KeyError as err:
        raise ConfigFormatError(f"The 'target_settings' section in the daq"
                                f" task {daq_label} is missing") from err
    try:
        target_power_on_config_path = target_config['power_on_default']
    except KeyError as err:
        raise ConfigFormatError(f"The 'power_on_default' key in the daq"
                                f" task {daq_label} is missing. It should"
                                "be a path to the power on default config"
                                " of the target") from err
    try:
        pwr_on_path = scan_file_dir /\
                target_power_on_config_path
        target_power_on_config = load_configuration(pwr_on_path)
    except FileNotFoundError as err:
        raise ConfigFormatError(f"The target power on configuration for"
                                f" {daq_label} could not be found") from err
    try:
        initial_config_path = target_config['initial_config']
    except KeyError as err:
        raise ConfigFormatError(f"The 'initial_config' key in the daq"
                                f" task {daq_label} is missing. It should"
                                "be a path to the configuration of the "
                                " target at the beginning of the measure"
                                "ments") from err
    try:
        init_cfg_path = scan_file_dir / initial_config_path
        target_initial_config = load_configuration(init_cfg_path.resolve())
    except FileNotFoundError as err:
        raise ConfigFormatError(f"The initial target config of {daq_label}"
                                " could not be found") from err

    try:
        columns = scan_config['data_columns']
        if not isinstance(columns, list):
            raise ConfigFormatError("The columns key must be a list of the "
                                    f" dataframe, columns is a {type(columns)}")
    except KeyError:
        columns = None

    # parse the scan dimensions
    scan_parameters = []
    if 'parameters' not in scan_config:
        raise ConfigFormatError("There needs to be a 'parameters' list"
                                f" in the scan entry {daq_label}")
    for scan_dimension in scan_config['parameters']:
        if ('key' not in scan_dimension and 'template' not in scan_dimension) or\
              (('range' not in scan_dimension) and
              ('values' not in scan_dimension)):
            raise ConfigFormatError("A range or a list of values along with a key"
                                    " must be specified"
                                    " for a scan dimension")
        if 'key' in scan_dimension and 'template' in scan_dimension:
            raise ConfigFormatError("'key' and 'template' are mutually exclusive");
        if 'key' in scan_dimension:
            if scan_dimension['key'] is None:
                continue
            scan_dim_key = scan_dimension['key']
            if not isinstance(scan_dim_key, list):
                raise ConfigFormatError("The key field in the parameter section"
                        f" of {daq_label} needs to be a list")
        if 'template' in scan_dimension:
            template = scan_dimension['template']
            try:
                _ = jinja2.Template(template)
            except jinja2.TemplateSyntaxError as e:
                raise ConfigFormatError("The template in"
                        " scan config {daq_label} is malformed") from e
            scan_dim_key = template

        step = 1
        start = None
        stop = None
        for key, value in scan_dimension.items():
            if key == 'range':
                for subkey, subval in value.items():
                    if subkey == 'step':
                        step = subval
                    if subkey == 'start':
                        start = subval
                    if subkey == 'stop':
                        stop = subval
                if start is None or stop is None:
                    raise ConfigFormatError(f"The Scan configuration {daq_label} is"
                                            " malformed. It misses either the start"
                                            " or stop entry for one of it's "
                                            " dimensions")
                scan_range = list(range(start, stop, step))

            elif key == 'values':
                if type(value)!=list:
                    raise ConfigFormatError(f"The Scan configuration {daq_label} is"
                                            " malformed. The values key should contain"
                                            " a list of values ")
                scan_range = value
        scan_parameters.append((scan_dim_key, scan_range))

    if len(scan_parameters) == 0:
        scan_parameters = [([], [0])]
    return {'parameters': scan_parameters,
            'calibration': calibration,
            'name': daq_label,
            'type': 'daq',
            'raw': raw_data,
            'target_power_on_default_config': target_power_on_config,
            'target_init_config': target_initial_config,
            'daq_system_config': daq_system_config,
            'daq_system_default_config': daq_system_default_config,
            'data_columns': columns}


def parse_analysis_config(config: dict) -> tuple:
    """
    parse the analysis configuration from a dict and return it in the
    same way that the parse_scan_config

    Returns
        Tuple containing the parameters relevant to the analysis given
        by the analysis_label along with the name of the daq task that
        provides the data for the analysis
        analysis_parameters: parameters of the analysis to be performed
        daq: daq task required by the analysis
        analysis_label: name of the analysis
    """
    try:
        event_mode = config['event_mode']
        if not isinstance(event_mode, bool):
            raise ConfigFormatError("The event_data field needs to contain a"
                                    " boolean")
    except KeyError:
        event_mode = False
    try:
        analysis_label = config['name']
    except KeyError as err:
        raise ConfigFormatError("The analysis configuration needs"
                                " a name field") from err
    try:
        analysis_object_name = config['python_module_name']
    except KeyError as err:
        raise ConfigFormatError("The analysis configuration needs"
                                " a python_module_name field") from err
    # parse the daq that is required for the analysis
    if 'daq' not in config.keys():
        raise ConfigFormatError("an analysis must specify a daq task")
    daq = config['daq']

    analysis_parameters = {}
    if 'parameters' in config.keys():
        if not isinstance(config['parameters'], dict):
            raise ConfigFormatError("The parameters of an analysis must be"
                                    " a dictionary")
        analysis_parameters = config['parameters']
    return {'parameters': analysis_parameters,
            'event_mode': event_mode,
            'daq': daq,
            'python_module': analysis_object_name,
            'name': analysis_label,
            'type': 'analysis'}


def parse_config_entry(entry: dict, path: str):
    """
    take care of calling the right function for the different configuration
    types calls the parse_scan_config and parse_analysis_config.
    """
    ptype = entry['type']
    if ptype.lower() == 'daq':
        return parse_scan_config(entry, path)
    if ptype.lower() == 'analysis':
        return parse_analysis_config(entry)
    raise ValueError("The type of a procedure must be"
                     " either daq or analysis")


def parse_config_file(config_path: str):
    """
    Parse the root config file and return a list of workflows and
    procedures that are fully parsed so that with them the analyses
    and scan can be properly initialized
    """
    path = Path(config_path).resolve()
    if not path.exists():
        raise FileNotFoundError("The filepath %s given does not exist"%path)
    procedures = []
    workflows = []
    config = load_configuration(path)
    if isinstance(config, dict):
        if 'libraries' in config.keys():
            # resolve the path to the library files taking
            # into account that both relative and absolute
            # paths can be specified in the libraries
            # section
            lib_root_dir = Path(os.path.dirname(path))
            other_paths = config['libraries']
            for lib_path in other_paths:
                if not str(lib_path).startswith('/'):
                    resolved_lib_path = lib_root_dir / lib_path
                else:
                    resolved_lib_path = Path(lib_path).resolve()
                other_procedures, other_workflows = parse_config_file(
                        resolved_lib_path)
                for procedure in other_procedures:
                    procedures.append(procedure)
                for workflow in other_workflows:
                    workflows.append(workflow)
        if 'workflows' in config.keys():
            pass
        # The root configuration file can also specify new
        # procedures, similar to libraries
        if 'procedures' in config.keys():
            if isinstance(config['procedures'], list):
                for procedure in config['procedures']:
                    ptype = procedure['type']
                    if ptype.lower() == 'daq':
                        procedures.append(parse_scan_config(procedure, path))
                    elif ptype.lower() == 'analysis':
                        procedures.append(parse_analysis_config(procedure))
                    else:
                        raise ValueError("The type of a procedure must be"
                                         " either daq or analysis")
            else:
                raise ValueError("procedures must be a list, not a dictionary")
    elif isinstance(config, list):
        for entry in config:
            procedures.append(parse_config_entry(entry, path))
    return procedures, workflows


#### TESTS ####
def test_load_config():
    test_fpath = Path('../../tests/configuration/scan_procedures.yaml')
    config = load_configuration(test_fpath)
    assert isinstance(config, list)
    assert isinstance(config[0], dict)


def test_update_dict():
    """
    test the update_dict function
    """
    test_fpath = Path('../../examples/daq_procedures.yaml')
    config = load_configuration(test_fpath)
    original = config[0]
    assert 'calibration' in original
    assert original['calibration'] == 'pedestal_calibration'
    update = {'calibration': 'vref_no_inv'}
    updated_dict = update_dict(original, update)
    assert updated_dict['calibration'] == 'vref_no_inv'
    original = {'this': [{'a': 1}, {'b': 2}, {'c': 3}, 3], 'that': 'what'}
    update = {'this': [{'a': 3}, {'b': 3}, {'c': 5}, 5]}
    expected_output = {'this': [{'a': 3}, {'b': 3}, {'c': 5}, 5],
                       'that': 'what'}
    updated_dict = update_dict(original, update)
    assert updated_dict == expected_output
    original = {'this': [{'a': 1}, {'b': 2}, {'c': 3}, 3], 'that': 'what'}
    update = {}
    expected_output = {'this': [{'a': 1}, {'b': 2}, {'c': 3}, 3],
                       'that': 'what'}
    updated_dict = update_dict(original, update)
    assert updated_dict == expected_output


def test_patch_generator():
    """
    test that the generate_patch_dict_from_key_tuple function
    """
    keys = [[['k1', 'k2', 'k3'], 'k4', 'k5'], ['daq', ['k1', 'k2', 'k3'], 'k4', 'k5']]
    value = 1
    expected_dict = [{'target' : {'k1': {'k4': {'k5': 1}},
                     'k2': {'k4': {'k5': 1}},
                     'k3': {'k4': {'k5': 1}}}},
                     {'daq' : {'k1': {'k4': {'k5': 1}},
                     'k2': {'k4': {'k5': 1}},
                     'k3': {'k4': {'k5': 1}}}}]
    for key, exd_dict in zip(keys, expected_dict):
        patch_dict = generate_patch(key, value)
        assert patch_dict == exd_dict


def test_diff_dict():
    test_dict_1 = {'a': 1, 'b': {'c': 2, 'f': 4}, 'e': 3}
    test_dict_2 = {'a': 2, 'b': {'c': 3, 'f': 4}, 'e': 3, 'g': 4}
    diff = {'a': 2, 'b': {'c': 3}, 'g': 4}
    result = diff_dict(test_dict_1, test_dict_2)
    assert result == diff
    result = diff_dict(test_dict_2, test_dict_1)
    assert result['b']['c'] == 2
    assert result['a'] == 1


def test_parse_scan_config():
    test_fpath = Path('../../examples/daq_procedures.yaml')
    scan_configs = []
    test_config = load_configuration(test_fpath)
    expected_config_keys = ['parameters', 'name', 'type',
                            'target_power_on_default_config',
                            'target_init_config',
                            'daq_system_config',
                            'daq_system_default_config',
                            'calibration', 'raw', 'data_columns']
    expected_column_keys = ['channel', 'chip', 'toa_mean', 'toa_stdd',
                            'channeltype']
    for scan in test_config:
        scan_configs.append(parse_scan_config(scan, test_fpath))
    assert len(scan_configs) == 3  # number of different scans
    test_config = scan_configs[0]
    for key in test_config.keys():
        assert key in expected_config_keys
    assert len(test_config.keys()) == len(expected_config_keys)
    assert test_config['name'] == 'injection_scan'
    assert len(test_config['parameters']) == 1
    assert test_config['parameters'][0][0][0] == 'level1_key1'
    for colname in test_config['data_columns']:
        assert colname in expected_column_keys
    assert len(test_config['data_columns']) == len(expected_column_keys)


def test_analysis_config():
    test_fpath = Path('../../examples/analysis_procedures.yaml')
    config = load_configuration(test_fpath)
    parsed_config = parse_analysis_config(config[0])
    assert parsed_config['daq'] == 'example_scan'
    assert isinstance(parsed_config['parameters'], dict)
    assert parsed_config['name'] == 'example_analysis'
    assert parsed_config['type'] == 'analysis'
    assert parsed_config['event_mode'] == False
    ana_params = parsed_config['parameters']
    assert ana_params['p1'] == 346045
    assert ana_params['p2'] == 45346
    assert ana_params['p3'] == 'string option'


def test_parse_config_entry():
    scan_test_config = Path('../../examples/daq_procedures.yaml')
    daq_system_test_config = Path(
            '../../tests/configuration/defaults/daq-system-config.yaml')
    daq_reference_config = load_configuration(daq_system_test_config)
    config = load_configuration(scan_test_config)
    names = ['injection_scan', 'timewalk_scan', 'example_scan']
    lengths = [1, 2, 1]
    expected_diffs = [{'server': {'NEvents': 4000}, 'client': {'hw_type': 'HD'}},
                      {'server': {'port': 7777, 'IdelayStep': 3},
                       'client': {'server_port': 7777}},
                      None]
    for entry, name, length, daq_diff in \
            zip(config, names, lengths, expected_diffs):
        parsed_config = parse_config_entry(entry, scan_test_config)
        assert parsed_config['name'] == name
        assert parsed_config['type'] == 'daq'
        assert len(parsed_config['parameters']) == length
        assert daq_diff == diff_dict(daq_reference_config,
                                    parsed_config['daq_system_config'])

def test_unfreeze():
    input_dict = FrozenOrderedDict({'a': 1, 'b': (1,2,3),
                                    'd': FrozenOrderedDict({'e': 4})})
    expected_dict = {'a': 1, 'b': [1,2,3],
                     'd': {'e': 4}}
    assert expected_dict == unfreeze(input_dict)

def test_parse_scan_config_template_handling():
    test_fpath = Path('../../tests/configuration/scan_procedures.yaml')
    config = load_configuration(test_fpath)
    template_test_config = config[1]
    test_scan_config = parse_scan_config(template_test_config, test_fpath)
    print(test_scan_config['parameters'])
    template = test_scan_config['parameters'][0][0]
    values = test_scan_config['parameters'][0][1]
    for value in values:
        patch = generate_patch(template, value)
        assert isinstance(patch['target'], dict)
        assert 'roc_s0' in patch['target']
        assert 'roc_s1' in patch['target']
        assert 'roc_s2' in patch['target']
        assert value in patch['target']['roc_s0']['ch']
        assert patch['target']['roc_s0']['ch'][value]['HighRange'] == 1



from setuptools import setup

setup(
    entry_points={
        'console_scripts': [
            'datenraffinerie = datenraffinerie.datenraffinerie:cli',
        ]
    },
    install_requires=[
        'Click',
        'luigi',
        'pandas',
        'matplotlib',
        'numpy',
        'scipy',
        'uproot',
        'pyyaml',
        'zmq',
        'pytest',
        'awkward',
        'tables',
        'h5py',
        'numba',
        'jinja2',
    ]
)
